#include <stdlib.h>
#include <stdio.h>
#include <string>
#include <vector>
#include <map>
#include <functional>
#include <iostream>
#include <fstream>
#include <sstream>

#define strnicmp _strnicmp
#define stricmp _stricmp

using std::function;
using std::string;
using std::vector;
using std::map;

const unsigned MEGADRIVE_FREQUENCY = 7670454; // Japan Mega Drive

template<typename T, int size> size_t numberof(T(&)[size]) { return size; }

// Warning: modifies the bitfield in-place
unsigned writeBits(unsigned& bitfield, unsigned bitNo, unsigned bitLength, unsigned writeValue) {
	if (writeValue >= 1U << bitLength) {
		printf("Value %x to write to bitfield %d:%d is too big, discarding.\n", writeValue, bitNo, bitLength);
		return bitfield;
	}

	unsigned mask = (1U << bitLength) - 1;
	bitfield &= ~(mask << bitNo);
	bitfield |= writeValue << bitNo;
	return bitfield;
}

template<typename T>
struct Property {
	function<T()> getter;
	operator T() { return getter(); }
};

struct Token {
	string text;
	int intContents; // Available for CHANNEL_NO, OPERATOR_NO, NUMBER
	double floatContents;
	enum Type {
		NEWLINE,
		END_OF_INPUT,
		FRAME,
		CHANNEL_NO,
		OPERATOR_NO,
		NUMBER,
		LFO,
		OFF,
		ON,
		KEYON,
		DETUNE,
		MULTIPLE,
		AMPLITUDE,
		RATE_SCALING,
		ATTACK_RATE,
		AMPLITUDE_MOD,
		DECAY_RATE,
		SUSTAIN_RATE,
		DECAY_LEVEL,
		RELEASE_RATE,
		// TODO SSEG-EG (0x90)
		FREQUENCY,
		OCTAVE,
		HZ,
		ALGORITHM,
		FEEDBACK,
		FMS,
		AMS,
		DB,
		PAN, PAN_DISABLED, PAN_LEFT, PAN_RIGHT, PAN_CENTER,
	} type;
	const string& programString;
	unsigned initialPosInProgram;

	Token(const string& programString, unsigned initialPosInProgram) : programString(programString), initialPosInProgram(initialPosInProgram), text(""), intContents(0) {}
};

static const struct {
	const char* text;
	Token::Type resultingType;
} tokenCorrespondanceTable[] = {
	{"frame", Token::Type::FRAME},
	{"frames", Token::Type::FRAME},
	{"lfo", Token::Type::LFO},
	{"off", Token::Type::OFF},
	{"on", Token::Type::OFF},
	{"keyon", Token::Type::KEYON},
	{"key-on", Token::Type::KEYON},
	{"detune", Token::Type::DETUNE},
	{"multiple", Token::Type::MULTIPLE},
	{"amplitude", Token::Type::AMPLITUDE},
	{"rate-scaling", Token::Type::RATE_SCALING},
	{"attack-rate", Token::Type::ATTACK_RATE},
	{"amplitude-mod", Token::Type::AMPLITUDE_MOD},
	{"amplitude-modulation", Token::Type::AMPLITUDE_MOD},
	{"decay-rate", Token::Type::DECAY_RATE},
	{"sustain-rate", Token::Type::SUSTAIN_RATE},
	{"decay-level", Token::Type::DECAY_LEVEL},
	{"release-rate", Token::Type::RELEASE_RATE},
	{"freq", Token::Type::FREQUENCY},
	{"frequency", Token::Type::FREQUENCY},
	{"hz", Token::Type::HZ},
	{"octave", Token::Type::OCTAVE},
	{"algo", Token::Type::ALGORITHM},
	{"algorithm", Token::Type::ALGORITHM},
	{"feedback", Token::Type::FEEDBACK},
	{"fms", Token::Type::FMS},
	{"ams", Token::Type::AMS},
	{"db", Token::Type::DB},
	{"pan", Token::Type::PAN},
	{"panning", Token::Type::PAN},
	{"disabled", Token::Type::PAN_DISABLED},
	{"left", Token::Type::PAN_LEFT},
	{"right", Token::Type::PAN_RIGHT},
	{"center", Token::Type::PAN_CENTER},
};

struct Error {
	string desc;
	string originalProgramString;
	unsigned positionInProgramString;

	Error(string desc, const Token& token) : desc(desc), originalProgramString(token.programString), positionInProgramString(token.initialPosInProgram) {}
};

struct Processor {
	string programStr;
	unsigned programStrIndex;
	map<unsigned, unsigned> pendingYmWrites;
	std::ofstream& outputFile;

	Processor(const char* program, std::ofstream& outputFile) : programStr(program), outputFile(outputFile) {
		programStrIndex = 0;
	}

	Property<char> currentChar = { [this]() { return programStrIndex < programStr.size() ? programStr[programStrIndex] : '\0'; } };

	char takeCurrentChar() {
		return programStr[programStrIndex++];
	}

	void positionToNextChar() {
		while (currentChar == ' ' || currentChar == '	') {
			programStrIndex += 1;
		}
	}

	bool isTokenStartChar(char c) {
		return c >= 'a' && c <= 'z' || c >= 'A' && c <= 'z';
	}

	bool isTokenChar(char c) {
		return isTokenStartChar(c) || c >= '0' && c <= '9';
	}

	bool isNumber(char c) {
		return c >= '0' && c <= '9' || c == '.';
	}

	void discardToken(Token tokenToReturnBefore) {
		programStrIndex = tokenToReturnBefore.initialPosInProgram;
	}

	void guessGeneralToken(Token& token) {
		if (!strnicmp(token.text.c_str(), "CH", 2) && token.text[2] >= '1' && token.text[2] <= '6') {
			token.type = Token::Type::CHANNEL_NO;
			token.intContents = token.text[2] - '1';
		}
		else if (!strnicmp(token.text.c_str(), "OP", 2) && token.text[2] >= '1' && token.text[2] <= '4') {
			token.type = Token::Type::OPERATOR_NO;
			token.intContents = token.text[2] - '1';
		}
		else {
			for (unsigned i = 0; i < numberof(tokenCorrespondanceTable); i++) {
				if (!stricmp(token.text.c_str(), tokenCorrespondanceTable[i].text)) {
					token.type = tokenCorrespondanceTable[i].resultingType;
					goto foundToken;
				}
			}
			throw Error("unknown token", token);
		foundToken:;
		}
	}

	Token nextToken() {
		positionToNextChar();
		Token result(programStr, programStrIndex);

		if (isTokenStartChar(currentChar)) {
			// General token
			do {
				result.text += takeCurrentChar();
			} while (isTokenChar(currentChar));
			guessGeneralToken(result);
		}
		else if (isNumber(currentChar)) {
			do {
				result.text += takeCurrentChar();
			} while (isNumber(currentChar));
			result.type = Token::Type::NUMBER;
			result.intContents = atoi(result.text.c_str());
			result.floatContents = atof(result.text.c_str());
		}
		else if (currentChar == '\n') {
			result.text = takeCurrentChar();
			result.type = Token::Type::NEWLINE;
		}
		else if (currentChar == '\0') {
			result.text = takeCurrentChar();
			result.type = Token::Type::END_OF_INPUT;
		}
		else {
			throw Error("unknown character", result);
		}

		return result;
	}

	Token::Type processLine();

	void process() {
		while (processLine() != Token::Type::END_OF_INPUT);
		writePendingFrame();
	}

	void enqueueYmWrite(unsigned address, unsigned data) {
		pendingYmWrites[address] = data;
	}

	void writePendingFrame() {
		for (auto it = pendingYmWrites.cbegin(); it != pendingYmWrites.cend(); ++it) {
			if (it->first < 0x100) outputFile << (uint8_t)1;
			else outputFile << (uint8_t)2;
			outputFile << (uint8_t)(it->first & 0xff);
			outputFile << (uint8_t)(it->second & 0xff);
		}

		outputFile << (uint8_t)0;
		pendingYmWrites.clear();
	}
};

struct LineProcessor {
	Processor& processor;
	int currentChannel, currentOperator;
	unsigned keyOnRegValue; // 0x28
	unsigned detuneMultipleRegValue[6][4] = { { 0 } }; // 0x30
	unsigned rateScalingAttackRateRegValue[6][4] = { { 0 } }; // 0x50
	unsigned amplitudeModDecayRateRegValue[6][4] = { { 0 } }; // 0x60
	unsigned decayLevelReleaseRateRegValue[6][4] = { { 0 } }; // 0x80
	unsigned frequencyLsbRegValue[6][4] = { { 0 } }; // 0xa0
	unsigned frequencyMsbRegValue[6][4] = { { 0 } }; // 0xa0
	unsigned algoFeedbackRegValue[6] = { { 0 } }; // 0xb0
	unsigned stereoLfoSensitivityRegValue[6] = { { 0 } }; // 0xb4

	LineProcessor() : processor(processor) {
		currentChannel = currentOperator = -1;
	}

	void expectChannelDefined(Token& token) {
		if (currentChannel < 0) throw Error("channel must be defined on this line", token);
	}

	void expectChannelAndOperatorDefined(Token& token) {
		expectChannelDefined(token);
		if (currentOperator < 0) throw Error("operator must be defined on this line", token);
	}

	unsigned expectUnsigned(Processor& processor) {
		Token next = processor.nextToken();
		if (next.type != Token::Type::NUMBER || next.intContents < 0) throw Error("positive integer expected", next);
		return unsigned(next.intContents);
	}

	unsigned expectUnsignedBits(const Token& next, unsigned maxBits) {
		if (next.type != Token::Type::NUMBER || next.intContents < 0) throw Error("positive integer expected", next);
		if (next.intContents >= 1 << maxBits) throw Error("number too big", next);
		return unsigned(next.intContents);
	}

	unsigned expectUnsignedBits(Processor& processor, unsigned maxBits) { return expectUnsignedBits(processor.nextToken(), maxBits); }

	unsigned makeAddressForChannel(unsigned baseAddress) {
		return currentChannel >= 3 ? (baseAddress + 0x100 + currentChannel - 3) : (baseAddress + currentChannel);
	}

	unsigned makeAddressForOperator(unsigned baseAddress) {
		unsigned channelNo = currentChannel, operatorNo = currentOperator;
		if (channelNo >= 3) {
			channelNo -= 3;
			baseAddress += 0x100;
		}
		return baseAddress + 4 * operatorNo + channelNo;
	}

	template<typename T>
	string arrayToString(T* array, unsigned arraySize) {
		std::ostringstream os;
		os << '{';
		for (unsigned i = 0; i < arraySize; i++) {
			if (i > 0) os << ", ";
			os << array[i];
		}
		os << '}';
		return os.str();
	}

	unsigned expectFloatInList(const Token& token, const double* list, unsigned listSize) {
		if (token.type != Token::Type::NUMBER) throw Error("number expected", token);
		for (unsigned i = 0; i < listSize; i++) {
			if (token.floatContents == list[i]) return i;
		}
		throw Error("unsupported numeric value, should be in " + arrayToString(list, listSize), token);
	}

	Token::Type process(Processor& processor) {
		const double fmsTable[] = { 0, 3.4, 6.7, 10, 14, 20, 40, 80 };
		const double amsTable[] = { 0, 1.4, 5.9, 11.8 };

		while (true) {
			Token token = processor.nextToken();
			//printf("ICI %d %s\n", token.type, token.text.c_str());

 			switch (token.type) {
			case Token::Type::NEWLINE:
			case Token::Type::END_OF_INPUT:
				return token.type;
			case Token::Type::FRAME: {
				unsigned numFrames = 1;
				Token times = processor.nextToken();
				if (times.type == Token::Type::NUMBER) numFrames = times.intContents;
				else processor.discardToken(times);
				while (numFrames-- > 0) processor.writePendingFrame();
				break;
			}

			case Token::Type::CHANNEL_NO:
				currentChannel = token.intContents;
				currentOperator = -1;
				break;
			case Token::Type::OPERATOR_NO:
				currentOperator = token.intContents;
				break;

			case Token::Type::LFO: {
				Token value = processor.nextToken();
				unsigned regValue = 0;
				if (value.type != Token::Type::OFF) {
					regValue = expectUnsignedBits(value, 3) | 1 << 3; // D3=LFO on
				}
				processor.enqueueYmWrite(0x22, regValue);
				break;
			}
			case Token::Type::KEYON: {
				Token channel = processor.nextToken();
				Token onOffForOperator[4] = {
					processor.nextToken(), processor.nextToken(), processor.nextToken(), processor.nextToken()
				};
				unsigned regValue = 0;

				if (channel.type != Token::Type::CHANNEL_NO) throw Error("expected channel", channel);
				for (unsigned i = 0; i < 4; i++) {
					if (onOffForOperator[i].type != Token::Type::ON && onOffForOperator[i].type != Token::Type::OFF) {
						throw Error("expected on/off for 4 operators", onOffForOperator[i]);
					}
					if (onOffForOperator[i].type != Token::Type::ON) {
						regValue |= 1 << (i + 4);
					}
				}
				if (channel.intContents < 3) regValue |= channel.intContents;
				else regValue |= channel.intContents + 1;
				processor.enqueueYmWrite(0x28, regValue);
				break;
			}
			case Token::Type::DETUNE:
				expectChannelAndOperatorDefined(token);
				processor.enqueueYmWrite(
					makeAddressForOperator(0x30),
					writeBits(detuneMultipleRegValue[currentChannel][currentOperator], 4, 3, expectUnsignedBits(processor, 3)));
				break;
			case Token::Type::MULTIPLE:
				expectChannelAndOperatorDefined(token);
				processor.enqueueYmWrite(
					makeAddressForOperator(0x30),
					writeBits(detuneMultipleRegValue[currentChannel][currentOperator], 0, 4, expectUnsignedBits(processor, 4)));
				break;
			case Token::Type::AMPLITUDE:
				expectChannelAndOperatorDefined(token);
				processor.enqueueYmWrite(makeAddressForOperator(0x40), expectUnsignedBits(processor, 7));
				break;
			case Token::Type::RATE_SCALING:
				expectChannelAndOperatorDefined(token);
				processor.enqueueYmWrite(
					makeAddressForOperator(0x50),
					writeBits(rateScalingAttackRateRegValue[currentChannel][currentOperator], 6, 2, expectUnsignedBits(processor, 2)));
				break;
			case Token::Type::ATTACK_RATE:
				expectChannelAndOperatorDefined(token);
				processor.enqueueYmWrite(
					makeAddressForOperator(0x50),
					writeBits(rateScalingAttackRateRegValue[currentChannel][currentOperator], 0, 5, expectUnsignedBits(processor, 6)));
				break;
			case Token::Type::AMPLITUDE_MOD:
				expectChannelAndOperatorDefined(token);
				processor.enqueueYmWrite(
					makeAddressForOperator(0x60),
					writeBits(amplitudeModDecayRateRegValue[currentChannel][currentOperator], 7, 1, expectUnsignedBits(processor, 1)));
				break;
			case Token::Type::DECAY_RATE:
				expectChannelAndOperatorDefined(token);
				processor.enqueueYmWrite(
					makeAddressForOperator(0x60),
					writeBits(amplitudeModDecayRateRegValue[currentChannel][currentOperator], 0, 5, expectUnsignedBits(processor, 5)));
				break;
			case Token::Type::SUSTAIN_RATE:
				expectChannelAndOperatorDefined(token);
				processor.enqueueYmWrite(makeAddressForOperator(0x70), expectUnsignedBits(processor, 5));
				break;
			case Token::Type::DECAY_LEVEL:
				expectChannelAndOperatorDefined(token);
				processor.enqueueYmWrite(
					makeAddressForOperator(0x80),
					writeBits(decayLevelReleaseRateRegValue[currentChannel][currentOperator], 4, 4, expectUnsignedBits(processor, 4)));
				break;
			case Token::Type::RELEASE_RATE:
				expectChannelAndOperatorDefined(token);
				processor.enqueueYmWrite(
					makeAddressForOperator(0x80),
					writeBits(decayLevelReleaseRateRegValue[currentChannel][currentOperator], 0, 4, expectUnsignedBits(processor, 4)));
				break;
			case Token::Type::FREQUENCY: {
				expectChannelDefined(token);
				unsigned freq = expectUnsignedBits(processor, 11);
				bool inHz = true;
				Token optionalHzAfter = processor.nextToken();
				if (optionalHzAfter.type != Token::Type::HZ) {
					processor.discardToken(optionalHzAfter);
					inHz = false;
				}
				if (inHz) {
					// F-number(note) = (144 * Fnote * 2^20 / Fm) / 2^(B-1)
					//freq = int(round((144 * freq * pow(2, 20) / MEGADRIVE_FREQUENCY) / pow(2, )
					throw Error("unsupported frequency in Hz, use a F-number and octave", token);
				}

				unsigned channel = currentChannel % 3, baseAddress = currentChannel < 3 ? 0xa0 : 0x1a0;
				if (channel < 2 || (channel == 2 && currentOperator <= 0)) { // not defined or operator 1
					processor.enqueueYmWrite(baseAddress + channel, writeBits(frequencyLsbRegValue[channel][0], 0, 8, freq & 0xff));
					processor.enqueueYmWrite(baseAddress + 4 + channel, writeBits(frequencyMsbRegValue[channel][0], 0, 3, freq >> 8));
				}
				else { // channel 3, op 2-4
					processor.enqueueYmWrite(
						baseAddress + 8 + currentOperator - 1,
						writeBits(frequencyLsbRegValue[channel][currentOperator], 0, 8, freq & 0xff));
					processor.enqueueYmWrite(
						baseAddress + 12 + currentOperator - 1,
						writeBits(frequencyMsbRegValue[channel][currentOperator], 0, 3, freq >> 8));
				}
				break;
			}
			case Token::Type::OCTAVE: {
				expectChannelDefined(token);
				unsigned octave = expectUnsignedBits(processor, 3);
				unsigned channel = currentChannel % 3, baseAddress = currentChannel < 3 ? 0xa0 : 0x1a0;
				if (octave >= 8) throw Error("octave number too large", token);
				if (channel < 2 || (channel == 2 && currentOperator <= 0)) { // not defined or operator 1
					processor.enqueueYmWrite(baseAddress + 4 + channel, writeBits(frequencyMsbRegValue[channel][0], 3, 3, octave));
				}
				else { // channel 3, op 2-4
					processor.enqueueYmWrite(
						baseAddress + 12 + currentOperator - 1,
						writeBits(frequencyMsbRegValue[channel][currentOperator], 3, 3, octave));
				}
				break;
			}
			case Token::Type::ALGORITHM:
				expectChannelDefined(token);
				processor.enqueueYmWrite(
					makeAddressForChannel(0xb0),
					writeBits(algoFeedbackRegValue[currentChannel], 0, 3, expectUnsignedBits(processor, 3)));
				break;
			case Token::Type::FEEDBACK:
				expectChannelDefined(token);
				processor.enqueueYmWrite(
					makeAddressForChannel(0xb0),
					writeBits(algoFeedbackRegValue[currentChannel], 3, 3, expectUnsignedBits(processor, 3)));
				break;
			case Token::Type::FMS:
				expectChannelDefined(token);
				processor.enqueueYmWrite(
					makeAddressForChannel(0xb4),
					writeBits(stereoLfoSensitivityRegValue[currentChannel], 0, 3, expectFloatInList(processor.nextToken(), fmsTable, numberof(fmsTable))));
				if (processor.nextToken().type != Token::Type::DB) throw Error("suffix value with dB required", token);
				break;
			case Token::Type::AMS:
				expectChannelDefined(token);
				processor.enqueueYmWrite(
					makeAddressForChannel(0xb4),
					writeBits(stereoLfoSensitivityRegValue[currentChannel], 4, 2, expectFloatInList(processor.nextToken(), amsTable, numberof(amsTable))));
				if (processor.nextToken().type != Token::Type::DB) throw Error("suffix value with dB required", token);
				break;
			case Token::Type::PAN:
				expectChannelDefined(token);
				Token next = processor.nextToken();
				unsigned twoBitValue = 0;
				if (next.type == Token::Type::PAN_CENTER) twoBitValue = 3;
				else if (next.type == Token::Type::PAN_LEFT) twoBitValue = 2;
				else if (next.type == Token::Type::PAN_RIGHT) twoBitValue = 1;
				else if (next.type == Token::Type::PAN_DISABLED) twoBitValue = 0;
				else throw Error("expected: center, left, right, disabled", next);
				processor.enqueueYmWrite(
					makeAddressForChannel(0xb4),
					writeBits(stereoLfoSensitivityRegValue[currentChannel], 6, 2, twoBitValue));
				break;
			};
		}

	}
};

inline Token::Type Processor::processLine() {
	LineProcessor lineProcessor;
	return lineProcessor.process(*this);
}

string readWholeFile(const char* fileName) {
	std::ifstream programFile(fileName);
	std::string str;
	programFile.seekg(0, std::ios::end);
	str.reserve(programFile.tellg());
	programFile.seekg(0, std::ios::beg);
	str.assign((std::istreambuf_iterator<char>(programFile)), std::istreambuf_iterator<char>());
	return str;
}

int main() {
	string programText = readWholeFile("input.txt");

	std::ifstream refFile;
	char gymData[428];
	refFile.open("reference.gym", std::ios::in | std::ios::binary);
	refFile.read(gymData, numberof(gymData));
	refFile.close();

	std::ofstream outputFile;
	outputFile.open("output.gym", std::ios::out | std::ios::binary);
	outputFile.write(gymData, numberof(gymData));

	try {
		Processor processor(programText.c_str(), outputFile);
		processor.process();
		outputFile.close();
	}
	catch (Error e) {
		printf("Error near %s\n%s", e.originalProgramString.c_str() + e.positionInProgramString, e.desc.c_str());
	}
	system("pause");
}

